import json
import pytest

from sweetpea._internal.logic import And, Or
from sweetpea._internal.backend import LowLevelRequest, BackendRequest


def test_low_level_request_validation():
    # Acceptable requests
    LowLevelRequest('EQ', 1, [1, 2, 3])
    LowLevelRequest('LT', 1, [1, 2, 3])
    LowLevelRequest('GT', 1, [1, 2, 3])

    # Invalid comparison
    with pytest.raises(ValueError):
        LowLevelRequest('bad', 1, [1, 2, 3])

    # Non-numeric k
    with pytest.raises(ValueError):
        LowLevelRequest('EQ', '5', [1, 2, 3])



import operator as op
import pytest

from itertools import permutations

from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition, Window
from sweetpea._internal.constraint import AtMostKInARow, Derivation, Reify
from sweetpea._internal.derivation_processor import DerivationProcessor
from sweetpea._internal.block import Block
from sweetpea import CrossBlock


color = Factor("color", ["red", "blue"])
text  = Factor("text",  ["red", "blue"])

con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

color_repeats_factor = Factor("repeated color?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])

text_repeats_factor = Factor("repeated text?", [
    DerivedLevel("yes", Transition(lambda texts: texts[0] == texts[-1], [text])),
    DerivedLevel("no",  Transition(lambda texts: texts[0] != texts[-1], [text]))
])

congruent_bookend = Factor("congruent bookend?", [
    DerivedLevel("yes", Window(lambda color, text: color == text, [color, text], 1, 3)),
    DerivedLevel("no",  Window(lambda color, text: color != text, [color, text], 1, 3))
])

design = [color, text, con_factor]
crossing = [color, text]
blk = CrossBlock(design, crossing, [Reify(con_factor)])


def two_con(i, n, t):
    return (i == "1" and n == "I" and t == "two") or \
        (i == "1" and n == "II" and t == "one") or \
        (i == "2" and n == "I" and t == "one") or \
        (i == "2" and n == "I" and t == "two") or \
        (i == "2" and n == "II" and t == "one") or \
        (i == "1" and n == "II" and t == "two")

def two_not_con(i, n, t):
    return not two_con(i, n, t)


def test_generate_derivations_should_raise_error_if_fn_doesnt_return_a_boolean():
    def local_eq(color, text):
            color == text # Notice the missing return stmt

    local_con_factor = Factor("congruent?", [
        DerivedLevel("con", WithinTrial(local_eq, [color, text])),
        DerivedLevel("inc", WithinTrial(lambda c, t: not local_eq(c, t), [color, text]))
    ])

    with pytest.raises(ValueError):
        CrossBlock([color, text, local_con_factor],
                   [color, text],
                   [Reify(local_con_factor)])

def test_generate_derivations_should_raise_error_if_some_factor_matches_multiple_levels():
    local_con_factor = Factor("congruent?", [
        DerivedLevel("con", WithinTrial(op.eq, [color, text])),
        DerivedLevel("inc", WithinTrial(op.eq, [color, text]))
    ])

    with pytest.raises(ValueError):
        CrossBlock([color, text, local_con_factor],
                   [color, text],
                   [Reify(local_con_factor)])

def test_generate_derivations_should_produce_warning_if_some_level_is_unreachable(capsys):
    local_con_factor = Factor("congruent?", [
        DerivedLevel("con", WithinTrial(op.eq, [color, text])),
        DerivedLevel("inc", WithinTrial(op.ne, [color, text])),
        DerivedLevel("dum", WithinTrial(lambda c, t: c=='green', [color, text]))
    ])
    block = CrossBlock([color, text, local_con_factor],
                       [color, text],
                       [Reify(local_con_factor)])
    block.show_errors()
    assert capsys.readouterr().out == "WARNING: No matches to the factor 'congruent?' predicate for level\n 'dum'.\n"

def test_generate_derivations_within_trial():
    assert DerivationProcessor.generate_derivations(blk) == [
        Derivation(4, [[0, 2], [1, 3]], con_factor),
        Derivation(5, [[0, 3], [1, 2]], con_factor)]

    integer = Factor("integer", ["1", "2"])
    numeral = Factor("numeral", ["I", "II"])
    text = Factor("text", ["one", "two"])

    twoConLevel = DerivedLevel("twoCon", WithinTrial(two_con, [integer, numeral, text]))
    twoNotConLevel = DerivedLevel("twoNotCon", WithinTrial(two_not_con, [integer, numeral, text]))
    two_con_factor = Factor("twoCon?", [twoConLevel, twoNotConLevel])

    one_two_design = [integer, numeral, text, two_con_factor]
    one_two_crossing = [integer, numeral, text]

    assert DerivationProcessor.generate_derivations(
        CrossBlock(one_two_design, one_two_crossing, list(map(Reify, one_two_design)))) == [
        Derivation(6, [[0, 2, 5], [0, 3, 4], [0, 3, 5], [1, 2, 4], [1, 2, 5], [1, 3, 4]], two_con_factor),
        Derivation(7, [[0, 2, 4], [1, 3, 5]], two_con_factor)]


@pytest.mark.parametrize('design',
    [[color, text, color_repeats_factor],
     [color, color_repeats_factor, text],
     [color_repeats_factor, color, text]])
def test_generate_derivations_transition(design):
    block = CrossBlock(design, [color, text], list(map(Reify, design)))

    assert DerivationProcessor.generate_derivations(block) == [
        Derivation(16, [[0, 4], [1, 5]], color_repeats_factor),
        Derivation(17, [[0, 5], [1, 4]], color_repeats_factor)
    ]


@pytest.mark.parametrize('design',
    [[color, text, color_repeats_factor, text_repeats_factor],
     [color, color_repeats_factor, text_repeats_factor, text],
     [color_repeats_factor, color, text_repeats_factor, text],
     [color_repeats_factor, text_repeats_factor, color, text]])
def test_generate_derivations_with_multiple_transitions(design):
    block = CrossBlock([color, text, color_repeats_factor, text_repeats_factor],
                       [color, text],
                       [Reify(color_repeats_factor), Reify(text_repeats_factor)])

    assert DerivationProcessor.generate_derivations(block) == [
        Derivation(16, [[0, 4], [1, 5]], color_repeats_factor),
        Derivation(17, [[0, 5], [1, 4]], color_repeats_factor),
        Derivation(22, [[2, 6], [3, 7]], text_repeats_factor),
        Derivation(23, [[2, 7], [3, 6]], text_repeats_factor)
    ]


def test_generate_derivations_with_window():
    block = CrossBlock([color, text, congruent_bookend], [color, text], [Reify(congruent_bookend)])

    assert DerivationProcessor.generate_derivations(block) == [
        Derivation(16, [[0, 2], [1, 3]], congruent_bookend),
        Derivation(17, [[0, 3], [1, 2]], congruent_bookend)
    ]


def test_generate_argument_list_with_within_trial():
    x_product = con_level.get_dependent_cross_product()

    assert DerivationProcessor.generate_argument_list(con_level, x_product[0]) == ['red', 'red']
    assert DerivationProcessor.generate_argument_list(con_level, x_product[1]) == ['red', 'blue']
    assert DerivationProcessor.generate_argument_list(con_level, x_product[2]) == ['blue', 'red']
    assert DerivationProcessor.generate_argument_list(con_level, x_product[3]) == ['blue', 'blue']


def test_generate_argument_list_with_transition():
    color_repeats_level = DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color]))
    x_product = color_repeats_level.get_dependent_cross_product()

    assert DerivationProcessor.generate_argument_list(color_repeats_level, x_product[0]) == [{-1: 'red', 0: 'red'}]
    assert DerivationProcessor.generate_argument_list(color_repeats_level, x_product[1]) == [{-1: 'red', 0: 'blue'}]
    assert DerivationProcessor.generate_argument_list(color_repeats_level, x_product[2]) == [{-1: 'blue', 0: 'red'}]
    assert DerivationProcessor.generate_argument_list(color_repeats_level, x_product[3]) == [{-1: 'blue', 0: 'blue'}]

    double_repeat_level = DerivedLevel("name", Transition(lambda colors, texts: True, [color, text]))
    x_product = double_repeat_level.get_dependent_cross_product()

    assert DerivationProcessor.generate_argument_list(color_repeats_level, x_product[0]) == [{-1: 'red', 0: 'red'}, {-1: 'red', 0: 'red'}]
    assert DerivationProcessor.generate_argument_list(color_repeats_level, x_product[1]) == [{-1: 'red', 0: 'red'}, {-1: 'red', 0: 'blue'}]
    assert DerivationProcessor.generate_argument_list(color_repeats_level, x_product[2]) == [{-1: 'red', 0: 'red'}, {-1: 'blue', 0: 'red'}]
    assert DerivationProcessor.generate_argument_list(color_repeats_level, x_product[3]) == [{-1: 'red', 0: 'red'}, {-1: 'blue', 0: 'blue'}]

    assert DerivationProcessor.generate_argument_list(color_repeats_level, x_product[15]) == [{-1: 'blue', 0: 'blue'}, {-1: 'blue', 0: 'blue'}]


def test_shift_window():
    assert DerivationProcessor.shift_window([[0, 0], [1, 1]], WithinTrial(lambda x: x, [color]), 0) == [[0, 0], [1, 1]]
    assert DerivationProcessor.shift_window([[0, 0], [1, 1]], Transition(lambda x: x, [color]), 4) == [[0, 4], [1, 5]]
    assert DerivationProcessor.shift_window([[0, 2, 4], [1, 3, 5]], Window(lambda x: x, [color], 2, 3), 6) == [[0, 8, 16], [1, 9, 17]]
    assert DerivationProcessor.shift_window([[1, 1, 1, 1], [2, 2, 2, 2]], Window(lambda x: x, [color], 2, 4), 10) == \
        [[1, 11, 21, 31], [2, 12, 22, 32]]



import operator as op
import pytest

from itertools import permutations

from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition, Window
from sweetpea._internal.constraint import AtMostKInARow, Derivation, Reify
from sweetpea._internal.derivation_processor import DerivationProcessor
from sweetpea._internal.block import Block
from sweetpea import CrossBlock


color  = Factor("color",  ["red", "blue"])
motion = Factor("motion", ["up", "down"])
task   = Factor("task",   ["color", "motion"])

def color_motion_congruent(color, motion):
    return ((color == "red") and (motion == "up")) or \
           ((color == "blue") and (motion == "down"))

def color_motion_incongruent(color, motion):
    return not color_motion_congruent(color, motion)

congruency = Factor("congruency", [
    DerivedLevel("con", WithinTrial(color_motion_congruent,   [color, motion])),
    DerivedLevel("inc", WithinTrial(color_motion_incongruent, [color, motion]))
])

def response_left(task, color, motion):
    return (task == "color"  and color  == "red") or \
        (task == "motion" and motion == "up")

def response_right(task, color, motion):
    return not response_left(task, color, motion)

response = Factor("response", [
    DerivedLevel("left",  WithinTrial(response_left,  [task, color, motion])),
    DerivedLevel("right", WithinTrial(response_right, [task, color, motion]))
])

response_transition = Factor("response transition", [
    DerivedLevel("repeat", Transition(lambda responses: responses[0] == responses[-1], [response])),
    DerivedLevel("switch", Transition(lambda responses: responses[0] != responses[-1], [response]))
])


def test_generate_derivations_with_transition_that_depends_on_derived_levels():
    block = CrossBlock([color, motion, task, response, response_transition],
                       [color, motion, task],
                       [Reify(response), Reify(response_transition)])
    derivations = DerivationProcessor.generate_derivations(block)

    assert Derivation(64, [[6, 14], [7, 15]], response_transition) in derivations
    assert Derivation(65, [[6, 15], [7, 14]], response_transition) in derivations


def test_generate_derivations_when_derived_factor_precedes_dependencies():
    block = CrossBlock([congruency, motion, color, task],
                       [color, motion, task],
                       [Reify(congruency)])
    derivations = DerivationProcessor.generate_derivations(block)

    assert Derivation(0, [[4, 2], [5, 3]], congruency) in derivations
    assert Derivation(1, [[4, 3], [5, 2]], congruency) in derivations



"""This module provides a simple base class for exceptions generated by running
a third-party tool.
"""


class ToolError(Exception):
    """An error raised when a command-line tool fails."""

    def __init__(self, returncode: int, stdout: str, stderr: str):
        message_lines = [str(returncode)]
        if stdout:
            message_lines.append("\n    stdout output captured below:\n" + stdout)
        if stderr:
            message_lines.append("\n    stderr output captured below:\n" + stderr)
        super().__init__(''.join(message_lines))



"""This module provides functionality for processing derivations."""


import operator as op

from typing import Any, Dict, List, Tuple, cast
from functools import reduce

from sweetpea._internal.primitive import Window, DerivedFactor, DerivedLevel, Level
from sweetpea._internal.block import Block
from sweetpea._internal.constraint import Derivation
from sweetpea._internal.iter import chunk_list, chunk_dict
from sweetpea._internal.beforestart import BeforeStart

class DerivationProcessor:

    @staticmethod
    def generate_derivations(block: Block) -> List[Derivation]:
        """Usage::

            >>> import operator as op
            >>> color = Factor("color", ["red", "blue"])
            >>> text  = Factor("text",  ["red", "blue"])
            >>> conLevel  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
            >>> incLevel  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
            >>> conFactor = Factor("congruent?", [conLevel, incLevel])
            >>> design = [color, text, conFactor]
            >>> crossing = [color, text]
            >>> block = fully_cross_block(design, crossing, [])
            >>> DerivationProcessor.generate_derivations(block)
            [Derivation(derivedIdx=4, dependentIdxs=[[0, 2], [1, 3]]), Derivation(derivedIdx=5, dependentIdxs=[[0, 3], [1, 2]])]

        In the example above, the indicies of the design are:

        ===  =============
        idx  level
        ===  =============
        0    color:red
        1    color:blue
        2    text:red
        3    text:blue
        4    conFactor:con
        5    conFactor:inc
        ===  =============

        So the tuple ``(4, [[0,2], [1,3]])`` represents the information that
        the derivedLevel con is true iff ``(color:red && text:red) ||
        (color:blue && text:blue)`` by pairing the relevant indices together.

        :rtype:
            returns a list of tuples. Each tuple is structured as:
            ``(index of the derived level, list of dependent levels)``
        """
        # We include implied factors in `derived_factors` so we check the mapping of levels,
        # but we'll only add to `accum` if the factor is non-implied
        derived_factors: List[DerivedFactor] = [factor for factor in block.design if isinstance(factor, DerivedFactor)]
        accum = []
        for factor in derived_factors:
            according_level: Dict[Tuple[Any, ...], DerivedLevel] = {}
            for level in factor.levels:
                cross_product: List[Tuple[Level, ...]] = level.get_dependent_cross_product()
                valid_tuples: List[Tuple[Level, ...]] = []
                for level_tuple in cross_product:
                    args = [(level.name if not isinstance(level, BeforeStart) else None) for level in level_tuple]
                    if level.window.width != 1:
                        # NOTE: mypy doesn't like this, but I'm not rewriting
                        #       it right now. Need to replace `chunk_dict` with
                        #       a better version.
                        args = list(chunk_dict(args, level.window.width))  # type: ignore
                    result = level.window.predicate(*args)
                    if not isinstance(result, bool):
                        raise ValueError(f"Expected derivation predicate to return bool; got {type(result)}.")
                    if result:
                        valid_tuples.append(level_tuple)
                        if level_tuple in according_level:
                            raise ValueError(f"Factor {factor.name} matches {according_level[level_tuple].name} and "
                                             f"{level.name} with assignment {args}.")
                        according_level[level_tuple] = level

                if not valid_tuples:
                    in_crossing = block.factor_in_crossing(factor)
                    # Solvers can go wrong if a crossing level is not even possible:
                    maybe_warning = "WARNING: " if (not in_crossing) or not block.require_complete_crossing else ""
                    maybe_crossing = "crossed " if in_crossing else ""
                    not_satisfiable = "not satisfiable" if block.require_complete_crossing else "incomplete"
                    conclusion = f", which means that the crossing is {not_satisfiable}" if in_crossing else ""
                    block.errors.add(f"{maybe_warning}No matches to the {maybe_crossing}factor"
                                     f" '{factor.name}' predicate for level\n '{level.name}'{conclusion}.")

                if factor in block.act_design:
                    # A `BeforeStart` in a tuple means that it should only be used for an early trial
                    # where the corresponding level is not yet available
                    valid_indices = [[(block.first_variable_for_level(level.factor, level) if not isinstance(level, BeforeStart) else level)
                                      for level in valid_tuple]
                                     for valid_tuple in valid_tuples]
                    shifted_indices = DerivationProcessor.shift_window(valid_indices,
                                                                       level.window,
                                                                       block.variables_per_trial())
                    level_index = block.first_variable_for_level(factor, level)
                    accum.append(Derivation(level_index, shifted_indices, factor))
        return accum

    @staticmethod
    def generate_argument_list(level: DerivedLevel, tup: Tuple[Level, ...]) -> List:
        # User-supplied string level names are the arguments for the user-supplied derivation functions
        level_strings = [level.name for level in tup]
        # For windows with a width of 1, we just pass the arguments directly, rather than putting them in lists.
        if level.window.width == 1:
            return level_strings
        else:
            return list(chunk_dict(level_strings, level.window.width))

    @staticmethod
    def shift_window(indices: List[List[object]],
                     window: Window,
                     trial_size: int
                     ) -> List[List[object]]:
        """This is a helper function that shifts the indices of
        :func:`.DerivationProcessor.generate_derivations`.

        E.g., if it's a ``Transition(op.eq, [color, color])`` (i.e., "repeat"
        color transition) then the indices for the levels of color would be
        like ``(0, 0)``, ``(1, 1)``, but actually, the window size for a
        transition is ``2``, so what we really want is the indices ``(0, 5)``,
        ``(1, 6)`` (assuming there are 4 levels in the design).

        So this helper function shifts over indices that were meant to be
        interpreted as being in a subsequent trial.
        """
        if window.width == 1:
            return indices

        # shifted_indices: List[List[int]] = []
        # shifted_sublists: List[List[int]] = []

        # factor_count = len(window.factors)

        # for index_list in indices:
        #     sublist_size = len(index_list) // factor_count
        #     sublists = chunk_list(index_list, sublist_size)
        #     shifted_sublists =

        shifted_idxs = cast(List[List[object]], [])
        shifted_sublists = cast(List[List[object]], [])
        argc = len(window.factors)
        for idx_list in indices:
            sublist_size = len(idx_list) // argc
            sublists = chunk_list(idx_list, sublist_size)
            shifted_sublists = []
            for idx_list in sublists:
                l = cast(List[object], [])
                for i, idx in enumerate(idx_list):
                    if isinstance(idx, BeforeStart):
                        l.append(BeforeStart(idx.ready_at+(len(idx_list) - i - 1)))
                    else:
                        l.append(cast(int, idx) + i * trial_size)
                shifted_sublists.append(l)
            shifted_idxs.append(list(reduce(op.add, shifted_sublists, [])))

        return shifted_idxs



from typing import List, cast, Tuple, Dict
from sweetpea._internal.weight import combination_weight
from sweetpea._internal.primitive import Level, Factor


def combinations_mismatched_weights(start: int, end: int, weight: int, crossing: List[Factor], sample: dict, or_less: bool):
    """returns how many mismatches of frequencies are in a sample against the expected crossing"""

    mismatch = 0
    combos = cast(Dict[Tuple[Level, ...], int], {})
    for t in range(start, end):
        key = tuple([sample[f][t] for f in crossing])
        combos[key] = combos.get(key, 0) + 1
    for combo, count in combos.items():
        delta = count - (combination_weight(combo) * weight)
        if or_less and (delta < 0):
            delta = 0
        mismatch += abs(delta)
    return mismatch



import operator as op
import pytest

from sweetpea._internal.primitive import Factor, DerivedLevel, Transition
from sweetpea._internal.constraint import Derivation, Reify
from sweetpea._internal.derivation_processor import DerivationProcessor
from sweetpea import CrossBlock


color_list = ["red", "green"]
color = Factor("color", color_list)
text  = Factor("text",  color_list)


def one_different(colors, texts):
    if (colors[0] == colors[-1]):
        return texts[0] != texts[-1]
    else:
        return texts[0] == texts[-1]

def other_different(colors, texts):
    return not one_different(colors, texts)

one_level   = DerivedLevel("one",  Transition(one_different, [color, text]))
other_level = DerivedLevel("other", Transition(other_different, [color, text]))
change = Factor("change", [one_level, other_level])

design      = [color, text, change]
crossing    = [color, text]
block       = CrossBlock(design, crossing, [Reify(change)])


# Encoding diagram minus 1
# -----------------------------------------------
# |   Trial |   color   |   text    |  change   |
# |       # | red green | red green | one other |
# -----------------------------------------------
# |       1 |  0    1   |  2    3   |           |
# |       2 |  4    5   |  6    7   | 16   17   |
# |       3 |  8    9   | 10   11   | 18   19   |
# |       4 | 12   13   | 14   15   | 20   21   |
# -----------------------------------------------
def test_generate_derivations():
    assert DerivationProcessor.generate_derivations(block) == [
        Derivation(16, [
            [0, 4, 2, 7],
            [0, 4, 3, 6],
            [0, 5, 2, 6],
            [0, 5, 3, 7],
            [1, 4, 2, 6],
            [1, 4, 3, 7],
            [1, 5, 2, 7],
            [1, 5, 3, 6]], change),
        Derivation(17, [
            [0, 4, 2, 6],
            [0, 4, 3, 7],
            [0, 5, 2, 7],
            [0, 5, 3, 6],
            [1, 4, 2, 7],
            [1, 4, 3, 6],
            [1, 5, 2, 6],
            [1, 5, 3, 7]], change)
    ]


def test_shift_window():
    assert DerivationProcessor.shift_window([[0, 0, 2, 3]], one_level.window, 4) == [[0, 4, 2, 7]]
    assert DerivationProcessor.shift_window([[1, 1, 3, 2]], one_level.window, 4) == [[1, 5, 3, 6]]

    assert DerivationProcessor.shift_window([[0, 0, 2, 2]], other_level.window, 4) == [[0, 4, 2, 6]]
    assert DerivationProcessor.shift_window([[0, 0, 3, 3]], other_level.window, 4) == [[0, 4, 3, 7]]



"""This module provides functionality for generating encoding diagrams."""


from itertools import repeat
from functools import reduce

from sweetpea._internal.primitive import DerivedFactor
from sweetpea._internal.block import Block
from sweetpea._internal.level import get_all_levels


def print_encoding_diagram(blk: Block) -> None:
    """Helper method to print a chart outlining the variable mappings, helpful
    for visualizing the formula space. For example, for the simple Stroop test,
    this would output::

        ----------------------------------------------
        |   Trial |  color   |   text   | congruent? |
        |       # | red blue | red blue |  con  inc  |
        ----------------------------------------------
        |       1 |  1   2   |  3   4   |   5    6   |
        |       2 |  7   8   |  9   10  |  11    12  |
        |       3 | 13   14  | 15   16  |  17    18  |
        |       4 | 19   20  | 21   22  |  23    24  |
        ----------------------------------------------
    """
    print(__generate_encoding_diagram(blk))


def __generate_encoding_diagram(blk: Block) -> str:
    diagram_str = ""

    design_size = blk.variables_per_trial()
    num_trials = blk.trials_per_sample()
    num_vars = blk.variables_per_sample()

    largest_number_len = len(str(num_vars))

    header_widths = []
    row_format_str = '| {:>7} |'
    for f in blk.design:
        # length of all levels concatenated for this factor
        level_names = [l.name for l in f.levels]
        level_name_widths = [max(largest_number_len, l) for l in list(map(len, level_names))]

        level_names_width = sum(level_name_widths) + len(level_names) - 1  # Extra length for spaces in between names.
        factor_header_width = max(len(f.name) if isinstance(f.name, str) else 0, level_names_width)
        header_widths.append(factor_header_width)

        # If the header is longer than the level widths combined, then they need to be lengthened.
        diff = factor_header_width - level_names_width
        if diff > 0:
            idx = 0
            while diff > 0:
                level_name_widths[idx] += 1
                idx += 1
                diff -= 1
                if idx >= len(level_name_widths):
                    idx = 0

        # While we're here, build up the row format str.
        row_format_str = reduce(lambda a, b: a + ' {{:^{}}}'.format(b), level_name_widths, row_format_str)
        row_format_str += ' |'

    header_format_str = reduce(lambda a, b: a + ' {{:^{}}} |'.format(b), header_widths, '| {:>7} |')
    factor_names = list(map(lambda f: f.name if isinstance(f.name, str) else f.name.name, blk.design))
    header_str = header_format_str.format(*["Trial"] + factor_names)
    row_width = len(header_str)

    # First line
    diagram_str += ('-' * row_width) + '\n'

    # Header
    diagram_str += header_str + '\n'

    # Level names
    all_level_names = [ln.name for (fn, ln) in get_all_levels(blk.design)]
    diagram_str += row_format_str.format(*['#'] + all_level_names) + '\n'

    # Separator
    diagram_str += ('-' * row_width) + '\n'

    # Variables
    for t in range(num_trials):
        args = [str(t + 1)]
        for f in blk.design:
            if f.applies_to_trial(t + 1):
                variables = [blk.first_variable_for_level(f, l) + 1 for l in f.levels]
                if isinstance(f, DerivedFactor) and f.has_complex_window:
                    def acc_width(w) -> int:
                        return w.width + (acc_width(w.factors[0].levels[0].window)-1 if w.factors[0].has_complex_window else 0)
                    width = acc_width(f.levels[0].window)
                    stride = f.levels[0].window.stride
                    stride_offset = (stride - 1) * int(t / stride)
                    offset = t - width + 1 - stride_offset
                    variables = list(map(lambda n: n + len(variables) * offset, variables))
                else:
                    variables = list(map(lambda n: n + design_size * t, variables))

                args += list(map(str, variables))
            else:
                args += list(repeat('', len(f.levels)))

        diagram_str += row_format_str.format(*args) + '\n'

    # Footer
    diagram_str += ('-' * row_width) + '\n'
    return diagram_str



import operator as op
import random
import numpy as np

from functools import reduce
from itertools import product
from math import factorial, ceil
from typing import List, cast, Tuple, Dict, Optional, Union, Any

from sweetpea._internal.block import Block
from sweetpea._internal.cross_block import CrossBlock
from sweetpea._internal.combinatorics import (
    n_choose_m,
    extract_components, compute_jth_permutation_prefix, compute_jth_combination,
    count_prefixes_of_permutations_with_copies, compute_jth_prefix_of_permutations_with_copies,
    PermutationMemo
)
from sweetpea._internal.design_partition import DesignPartitions
from sweetpea._internal.logic import And
from sweetpea._internal.primitive import SimpleLevel, Factor, DerivedFactor, Level
from sweetpea._internal.sampling_strategy.base import Gen, SamplingResult
from sweetpea._internal.constraint import Exclude, _KInARow, ExactlyKInARow, AtMostKInARow
from sweetpea._internal.iter import chunk
from sweetpea._internal.weight import combination_weight
from sweetpea._internal.check_mismatch import combinations_mismatched_weights


class RandomGen(Gen):
    """This strategy represents the ideal. Valid sequences are uniformly
    sampled via a bijection from natural numbers to valid trial sequences.

    Complex windows and counting constrants are handled by rejection sampling.
    """

    def __str__(self):
        return RandomGen.class_name()

    @staticmethod
    def class_name():
        return 'RandomGen'

    @staticmethod
    def sample(block: Block, sample_count: int) -> SamplingResult:
        return RandomGen.__sample(block, sample_count, 0)

    def __init__(self, acceptable_error):
        self.acceptable_error = acceptable_error

    def sample_object(self, block: Block, sample_count: int) -> SamplingResult:
        return RandomGen.__sample(block, sample_count, self.acceptable_error)

    @staticmethod
    def __sample(block: Block, sample_count: int, acceptable_error: int) -> SamplingResult:
        # 1. Validate the block.
        RandomGen.__validate(block)
        metrics = {}

        if block.show_errors():
            return SamplingResult([], {})

        # 2. Count how many solutions there are. The enumerator will note
        # the crossing size and minimum-trial request, and it will be prepared
        # to generate runs of a crossing-size length or "leftover" length.
        print("Counting possible configurations...")
        enumerator = UCSolutionEnumerator(cast(CrossBlock, block))
        metrics['solution_count'] = enumerator.solution_count()

        if (enumerator.solution_count() == 0):
            return SamplingResult([], metrics)

        crossing_size = enumerator.crossing_size # includes crossing weight

        # 3. Generate samples.
        print("Generating samples...")
        metrics['rejections'] = []
        sampled = 0
        rejected = 0
        total_rejected = 0
        trials_per_run = block.trials_per_sample()
        rounds_per_run = (trials_per_run - enumerator._preamble_size) // crossing_size
        leftover = (trials_per_run - enumerator._preamble_size) % crossing_size
        samples = cast(List[dict], [])
        used_keys = cast(Dict[Tuple[int, ...], bool], {})
        possible_keys = (enumerator.preamble_solution_count()
                         * pow(enumerator.solution_count(), rounds_per_run)
                         * enumerator.leftover_solution_count())
        while sampled < sample_count:
            if len(used_keys) == possible_keys:
                break

            solution_variabless = enumerator.generate_random_samples(rounds_per_run, leftover, used_keys)
            used_keys[enumerator.extract_sequence_key(solution_variabless)] = True

            # Combine randomly selected crossing-sized runs plus a leftover-sized run
            # into one complete run with the requested number of trials.
            run = solution_variabless[0][1]  # the preamble solution, possibly empty
            for round in range(0, rounds_per_run + (1 if leftover > 0 else 0)):
                run = RandomGen.__combine_round(run, solution_variabless[round + 1][1])

            run = enumerator.fill_in_nonpreamble_uncrossed_derived(run, trials_per_run)

            if RandomGen.__are_constraints_violated(cast(CrossBlock, block), run, enumerator,
                                                    rounds_per_run, leftover,
                                                    acceptable_error):
                rejected += 1
                if rejected % 10000 == 0:
                    if len(samples) > 0:
                        accepts = f", accepted {len(samples)}"
                    else:
                        accepts = ""
                    n = total_rejected + rejected
                    print(f"Rejected {n} candidates so far (out of {possible_keys} choices){accepts}")
                continue

            metrics['rejections'].append(rejected)
            total_rejected += rejected
            rejected = 0
            sampled += 1

            samples.append(enumerator.factors_and_levels_to_names(run))

        metrics['sample_count'] = sample_count
        metrics['total_rejected'] = total_rejected
        metrics['avg_rejected'] = total_rejected / sample_count
        if (total_rejected > 10000):
            print("")

        return SamplingResult(samples, metrics)

    @staticmethod
    def __are_constraints_violated(block: CrossBlock, sample: dict, enumerator: 'UCSolutionEnumerator',
                                   rounds_per_run: int, leftover: int,
                                   acceptable_error: int) -> bool:
        for ct in block.constraints:
            if not ct.potential_sample_conforms(sample, block):
                return True
        if enumerator.has_crossed_complex_derived_factors or len(block.crossings) > 1:
            # Check whether the sample achieves each crossing in the run
            bad = 0
            run_length = enumerator._preamble_size + (rounds_per_run * enumerator.crossing_size) + leftover
            for i, c in enumerate(block.crossings):
                if enumerator.has_crossed_complex_derived_factors or i > 0:
                    start = enumerator.preamble_sizes[i]
                    c_weight = enumerator.crossing_weights[i]
                    c_crossing_size = enumerator.crossing_sizes[i] * c_weight
                    c_rounds_per_run = (run_length - start) // c_crossing_size
                    c_leftover = (run_length - start) % c_crossing_size
                    # There might be a little room for improvement here by bailing
                    # out within combinations_mismatched_weights() as soon as the
                    # acceptable error is exceeded, but probably it doesn't matter.
                    for round in range(c_rounds_per_run):
                        # if not conds_match_weights(start, start + crossing_size, True):
                        #     return True
                        bad += combinations_mismatched_weights(start, start + c_crossing_size, c_weight, c, sample, False)
                        if bad > acceptable_error:
                            return True
                        start += c_crossing_size
                    if c_leftover > 0:
                        # if not conds_match_weights(start, start + c_leftover, True):
                        #     return True
                        bad += combinations_mismatched_weights(start, start + c_leftover, c_weight, c, sample, True)
                        if bad > acceptable_error:
                            return True
        return False

    @staticmethod
    def __validate(block: Block) -> None:
        # Triggers checks within `block`:
        block.trials_per_sample()

    @staticmethod
    def __combine_round(run: dict, round: dict) -> dict:
        if len(run) == 0:
            return round
        else:
            new_run = run.copy()
            for key in round:
                new_run[key] = run[key] + round[key]
            return new_run


Components = Tuple[int, Tuple[int, ...], Tuple[int, ...]]


class RandomComponentsShape():
    def __init__(self) -> None:
        self.crossings_shape = 0
        self.combinations_shapes = cast(List[int], [])
        self.independent_shapes = cast(List[int], [])


"""
Given a fully crossed block with no complex windows, this class stores the data structures and logic for enumerating
valid trial sequences in the design.
"""


class UCSolutionEnumerator():

    def __init__(self, block: CrossBlock) -> None:
        self._block = block
        self._partitions = DesignPartitions(block)
        self._crossing_instances = self.__generate_crossing_instances()
        c_weight = block.crossing_weight(block.crossings[0])
        self._crossing_weights = [combination_weight(tuple(c.values())) * c_weight
                                  for c in self._crossing_instances]
        self._crossing_is_unweighted = all([w == 1 for w in self._crossing_weights])
        self.noncomplex_crossing_size = sum(self._crossing_weights)
        self._source_combinations = self.__generate_source_combinations()
        self.__complex_crossing_instances = self.__count_complex_crossing_instances()
        # This crossing size includes the crossing weight (in addition to level weights)
        self.crossing_size = self.noncomplex_crossing_size * self.__complex_crossing_instances
        self._ind_factor_levels = cast(List[Tuple[Factor, List[SimpleLevel]]],
                                       [])  # Will be populated by solution counting
        self._basic_factor_levels = cast(List[Tuple[Factor, List[SimpleLevel]]],
                                         [])  # Will be populated by solution counting
        m = self.__complex_crossing_instances
        self._m_or_counters = cast(Union[int, List[int]],
                                   m if self._crossing_is_unweighted else [w * m for w in self._crossing_weights])

        self._sorted_derived_factors = self._partitions.get_derived_factors()
        self._sorted_derived_factors.sort(key=lambda f: f._get_depth())

        uc_deriv_and_complex = self._partitions.get_uncrossed_derived_and_complex_derived_factors()
        self._sorted_uncrossed_derived_and_complex_derived = uc_deriv_and_complex
        self._sorted_uncrossed_derived_and_complex_derived.sort(key=lambda f: f._get_depth())

        self.has_crossed_complex_derived_factors = (self.__complex_crossing_instances > 1)

        # Maintains a lookup for valid source combinations for a permutation.
        # Example [[2, 3], ...] means that for permutation 0, indices 2 and 3 in the source combinations
        # list are allowed.
        self._valid_source_combinations_indices = cast(List[List[int]], [])  # Will be populated by solution counting

        # For multiple crossings, we'll need the size of each crossing and how much to consider the
        # preamble of each; note that `self.crossing_sizes` does *not* include the crossing weight:
        self.crossing_sizes = [block.crossing_size(c) for c in block.crossings]
        self.preamble_sizes = [block.preamble_size(c) for c in block.crossings]
        self.crossing_weights = [block.crossing_weight(c) for c in block.crossings]
        # Check that calculations from two sources agree:
        assert self.crossing_sizes[0] * self.crossing_weights[0] == self.crossing_size

        noncomplex_crossing_size = self.noncomplex_crossing_size
        preamble_size = self.preamble_sizes[0]
        self._preamble_size = preamble_size;

        # Call `__count_solutions` after everything else is set up.
        self._components_shape = RandomComponentsShape()
        self._pmemo = PermutationMemo()
        self._solution_count = self.__count_solutions(self.crossing_size,
                                                      self._components_shape,
                                                      self._pmemo,
                                                      self._valid_source_combinations_indices,
                                                      self._ind_factor_levels)
        self._leftover_components_shape = RandomComponentsShape()
        self._leftover_solution_count = 1
        self._leftover_pmemo = PermutationMemo()
        leftover = (block.trials_per_sample() - preamble_size) % self.crossing_size;
        if (leftover != 0):
            self._leftover_solution_count = self.__count_solutions(leftover,
                                                                   self._leftover_components_shape,
                                                                   self._leftover_pmemo,
                                                                   None,
                                                                   None)

        # How many extra trials do we need to generate by randomly
        # selecting among non-derived factors, to support derived
        # factors with complex windows?
        self._preamble_solution_count = self.__count_preamble_solutions()

    def solution_count(self):
        return self._solution_count

    def leftover_solution_count(self):
        return self._leftover_solution_count

    def preamble_solution_count(self):
        return self._preamble_solution_count

    def crossing_instances_count(self):
        return len(self._crossing_instances)

    def generate_random_samples(self, n: int, leftover: int, sampled: Dict[Tuple[int, ...], bool]) -> List[
        Tuple[int, dict]]:
        # Select a collection of random numbers, each from the range of solutions.
        # If `leftover` is not zero, then tack on one more sequence that is shorter
        # than the crossing size.
        # Then, make sure we haven't already picked the same sequence according to `sampled`.
        # This rejection-based approach is probably ok for realistic experiments,
        # where we're unlikely to want a number of samples close to the number of solutions
        # at the same time that there are a lot of solutions.
        choice = cast(List[Any],
                      (random.randrange(0, self._preamble_solution_count),
                       tuple([self.random_components(self._components_shape, self.crossing_size, 0) for i in range(n)]),
                       self.random_components(self._leftover_components_shape, leftover,
                                              leftover) if leftover > 0 else 0))
        while tuple([choice[0]] + list(choice[1]) + ([choice[2]] if leftover > 0 else [])) in sampled:
            choice = cast(List[Any],
                          (random.randrange(0, self._preamble_solution_count),
                           tuple([self.random_components(self._components_shape, self.crossing_size, 0) for i in
                                  range(n)]),
                           self.random_components(self._leftover_components_shape, leftover,
                                                  leftover) if leftover > 0 else 0))
        return ([(choice[0], self.generate_preamble_sample(choice[0]))]
                + [(components, self.generate_sample_from_components(components)) for components in choice[1]]
                + ([(choice[2],
                     self.generate_leftover_sample(choice[2], leftover))] if leftover > 0 else []))

    def random_components(self, components_shape: RandomComponentsShape, trial_count: int, leftover: int) -> Components:
        crossing_permutation_index = random.randrange(0, components_shape.crossings_shape)
        if trial_count == len(self._crossing_instances) and self._crossing_is_unweighted:
            source_combination_indices = tuple([random.randrange(0, len)
                                                for len in components_shape.combinations_shapes])
        else:
            # The indicies for source combination depend on the chosen permutation
            permutation_indices = self.jth_permutation_indices(len(self._crossing_instances),
                                                               self.crossing_size if leftover == 0 else leftover,
                                                               crossing_permutation_index,
                                                               self._pmemo if leftover == 0 else self._leftover_pmemo)
            indices = []
            for p in permutation_indices:
                indices.append(random.randrange(0, components_shape.combinations_shapes[p]))
            source_combination_indices = tuple(indices)
        independent_factor_combination_indices = tuple([random.randrange(0, len)
                                                        for len in components_shape.independent_shapes])
        return (crossing_permutation_index,
                source_combination_indices,
                independent_factor_combination_indices)

    def extract_sequence_key(self, solution_variabless: List[Tuple[int, dict]]) -> Tuple[int, ...]:
        return tuple(map(lambda sv: sv[0], solution_variabless))

    def generate_sample_from_components(self, components: Components) -> dict:
        trial_values = self.generate_trial_values(components, self.crossing_size, len(self._crossing_instances),
                                                  self._pmemo)
        return self._trial_values_to_experiment(trial_values)

    def generate_sample(self, sequence_number: int) -> dict:
        """Used for tests where a single random sequence number is provided, instead of random components."""
        shapes = ([self._components_shape.crossings_shape]
                  + self._components_shape.combinations_shapes
                  + self._components_shape.independent_shapes)
        flat_components = extract_components(shapes, sequence_number)
        components = (flat_components[0],
                      tuple(flat_components[1:1 + self.crossing_size]),
                      tuple(flat_components[1 + self.crossing_size:]))
        return self.generate_sample_from_components(components)

    def generate_leftover_sample(self, components: Components, leftover: int) -> dict:
        trial_values = self.generate_trial_values(components, leftover, len(self._crossing_instances),
                                                  self._leftover_pmemo)
        return self._trial_values_to_experiment(trial_values)

    def _trial_values_to_experiment(self, trial_values: List[dict]) -> dict:
        experiment = cast(dict, {})
        for trial_number, trial_value in enumerate(trial_values):
            for factor, level in trial_value.items():
                if factor not in experiment:
                    experiment[factor] = []
                experiment[factor].append(level)
        return experiment

    def factors_and_levels_to_names(self, experiment: dict) -> dict:
        new_experiment = cast(dict, {})
        for k in experiment:
            new_experiment[k.name] = [(l.name if l else "") for l in experiment[k]]
        return new_experiment

    # Used for an acceptance test:
    def generate_solution_variables(self) -> List[int]:
        components = self.random_components(self._components_shape, self.crossing_size, 0)
        trial_values = self.generate_trial_values(components, len(self._crossing_instances),
                                                  len(self._crossing_instances),
                                                  PermutationMemo())

        solution = cast(List[int], [])
        # Convert to variable encoding for SAT checking
        for trial_number, trial_value in enumerate(trial_values):
            for factor, level in trial_value.items():
                solution.append(self._block.get_variable(trial_number + 1, (factor, level)))
        solution.sort()
        return solution

    def generate_trial_values(self, components: Components, trial_count: int, crossing_size: int,
                              pmemo: PermutationMemo) -> List[dict]:
        # Component pieces:
        #    The 0th component is always the permutation index for crossing combinations.
        #    The 1st component is a list of source combination indices. The indices are
        #    organized in two possible ways:
        #        1. One per crossing combination; used in the case that there's a 1-to-1 mapping
        #           between crossing combinations and trials.
        #        2. One per trial; used in the case that combinations have weights (so a
        #           combinations can appear multiple times) or we're generating leftover trials
        #           (so some combinations may be missing).
        #    The 2nd component is a list of combination indices for independent basic factors.

        # Generate the inversion sequence for the selected permutation number.
        # Use the inversion sequence to construct the permutation.
        permutation_indices = self.jth_permutation_indices(crossing_size, trial_count, components[0], pmemo)

        permutation = list(map(lambda i: self._crossing_instances[i], permutation_indices))

        # Generate the source combinations for the selected sequence.
        source_combinations = cast(List[dict], [])
        for i, p in enumerate(permutation_indices):
            if trial_count == len(self._crossing_instances) and self._crossing_is_unweighted:
                component_for_p = components[1][p]
            else:
                component_for_p = components[1][i]
            source_combination_index_for_component = self._valid_source_combinations_indices[p][component_for_p]
            source_combinations.append(self._source_combinations[source_combination_index_for_component])

        # Generate the combinations for independent basic factors
        independent_factor_combinations = cast(List[dict], [{}] * trial_count)
        if self._ind_factor_levels:
            for j, (fi, levels) in enumerate(self._ind_factor_levels):
                independent_combination_idx = components[2][j]
                combo = compute_jth_combination(trial_count, len(levels), independent_combination_idx)
                for i in range(trial_count):
                    if not independent_factor_combinations[i]:
                        independent_factor_combinations[i] = {fi: levels[combo[i]]}
                        continue
                    independent_factor_combinations[i][fi] = levels[combo[i]]

        # 5. Merge the selected levels gathered so far to facilitate computing the uncrossed derived factor levels.
        trial_values = cast(List[dict], [{}] * trial_count)
        for t in range(trial_count):
            trial_values[t] = {**permutation[t], **source_combinations[t], **independent_factor_combinations[t]}

        return trial_values

    def jth_permutation_indices(self, crossing_size: int, trial_count: int, component: int, pmemo: PermutationMemo):
        if self.__complex_crossing_instances == 1 and self._crossing_is_unweighted:
            return compute_jth_permutation_prefix(crossing_size, trial_count, component)
        else:
            return compute_jth_prefix_of_permutations_with_copies(crossing_size, self._m_or_counters,
                                                                  trial_count, component, pmemo)

    def generate_preamble_sample(self, sequence_number: int) -> dict:
        run = cast(dict, {})
        if self._preamble_size == 0:
            assert sequence_number == 0
            return run
        for f, levels in self._basic_factor_levels:
            trials = []
            for _ in range(self._preamble_size):
                n = sequence_number % len(levels)
                sequence_number = sequence_number // len(levels)
                trials.append(levels[n])
            run[f] = trials
        return self._fill_in_derived(run, self._sorted_derived_factors, 0, self._preamble_size)

    def fill_in_nonpreamble_uncrossed_derived(self, run: dict, trials_per_run: int) -> dict:
        # Note: "uncrossed" includes crossed derived that have complex windows
        return self._fill_in_derived(run, self._sorted_uncrossed_derived_and_complex_derived, self._preamble_size,
                                     trials_per_run)

    def _fill_in_derived(self, run: dict, sorted_factors: List[DerivedFactor], start: int, end: int) -> dict:
        for df in sorted_factors:
            if start > 0:
                trials = run[df][:start]
            else:
                trials = []
            for i in range(start, end):
                if df.applies_to_trial(i + 1):
                    trials.append(df.select_level_for_sample(i, run))
                else:
                    trials.append(None)
            run[df] = trials
        return run

    """
    Generates all the crossings, indexed by factor name for easy lookup later.
    [
        {'factor': 'level', 'factor': 'level', ...},
        ...
    ]
    """

    def __generate_crossing_instances(self) -> List[dict]:
        crossing = self._partitions.get_crossed_noncomplex_factors()
        level_lists = [list(f.levels) for f in crossing]
        crossings = [{crossing[i]: level for i, level in enumerate(levels)} for levels in product(*level_lists)]
        return list(filter(lambda c: not self._block.is_excluded_or_inconsistent_combination(c), crossings))

    def __count_complex_crossing_instances(self) -> int:
        crossing = self._partitions.get_crossed_complex_factors()
        if len(crossing) == 0:
            return 1
        else:
            level_lists = [list(f.levels) for f in crossing]
            crossings = [{crossing[i]: level for i, level in enumerate(levels)} for levels in product(*level_lists)]
            return sum([(0 if self._block.is_excluded_combination(c) else combination_weight(tuple(c.values())))
                        for c in crossings])

    def __generate_source_combinations(self) -> List[dict]:
        ubs = self._partitions.get_uncrossed_basic_source_factors()
        level_lists = [list(f.levels) for f in ubs]
        return [{ubs[i]: level for i, level in enumerate(levels)} for levels in product(*level_lists)]

    def __count_solutions(self,
                          first_n: int,
                          components_shape: RandomComponentsShape,
                          pmemo: PermutationMemo,
                          valid_source_combinations_indices: Optional[List[List[int]]],
                          ind_factor_levels: Optional[List[Tuple[Factor, List[SimpleLevel]]]]):
        """Returns the number of solutions for a single round. When
        minimum_trials increases the number of rounds, then we have
        roughly a power of this result. The first_n argument is
        between 1 and the crossing size, and it indicates how many of
        the possible crossing-size trials we'll keep in a run. If we
        need more trials than the crossing size, we'll pick multiple
        crossing-size runs and concatenate them.
        """
        ##############################################################
        # Permutations of crossing instances
        #
        # In the case of derived factors in the crossing, we consider only
        # those with simple windows. Derived factors with simple windows
        # can be combined with all other combinations, otherwise the design
        # has no solutions, anyway (and we'll detect that by discovering zero
        # allowed combinations in "Uncrossed Dependent Factors" below).
        # Derived factors with complex windows, meanwhile, mean that we
        # have to use rejection sampling, and so it's ok for the
        # "Uncrossed Dependent Factors" step to over-approximate the allowed
        # combinations.
        q = len(self._crossing_instances)
        m = self.__complex_crossing_instances
        n = q * m
        if m == 1 and self._crossing_is_unweighted:
            permutations = factorial(n)
            # We want only the first_n of the trials, so permutations of the
            # remaining trials would all count the same:
            if first_n != n:
                permutations = permutations // factorial(n - first_n)
        else:
            # We need n = q*m trials (well, the first_n of those). But we don't
            # want to distinguish among combinations where the noncomplex crossing
            # component (with m choices of each) is the same. More generally, we
            # weight each combination by a multple of m
            permutations = count_prefixes_of_permutations_with_copies(q, self._m_or_counters, first_n, pmemo)

        solution_count = permutations
        components_shape.crossings_shape = permutations

        ##############################################################
        # Uncrossed Dependent Factors
        #
        # For each combination of crossed factors, determine the possible
        # completions with uncrossed basic factors. We consider derived
        # factors only in the way that they limit completions via
        # uncrossed basic factors.
        level_combinations = self.__generate_source_combinations()

        # Keep only allowed combos for each permutation
        for ci in self._crossing_instances:
            sc_indices = list(range(len(self._source_combinations)))
            for sc_idx, sc in enumerate(self._source_combinations):
                # Apply the derivation fn for each DF in the crossing for this instance and make sure it
                # return true for this level combination. If it doesn't, then remove this combination.
                # The "crossing" here doesn't include crossed derived factors with complex windows,
                # though, so we can only perform this filtering for a derived factor that does
                # not depend on a complex-window derived factor.
                merged_levels = {**ci, **sc}
                for df in self._partitions.get_crossed_noncomplex_derived_factors():
                    # Not yet separating complex:
                    # assert not df.has_complex_window
                    if not df.has_complex_window:
                        w = merged_levels[df].window
                        if not w.predicate(*[(merged_levels[f]).name for f in w.factors]):
                            sc_indices.remove(sc_idx)

            components_shape.combinations_shapes.append(len(sc_indices))
            if isinstance(valid_source_combinations_indices, list):
                valid_source_combinations_indices.append(sc_indices)

        # When we're not generating a leftover or weighted sequence, every crossing will appear
        # once somewhere in the sequence, so we get to pick from all possible completions
        # of all combinations; in that case, we can just multiply the new segment
        # lengths into `solution_count`. Otherwise, we need to consider every choice of
        # `first_n` crossing combinations, and then multiply the
        if first_n == len(self._crossing_instances) and self._crossing_is_unweighted:
            solution_count *= reduce(op.mul, components_shape.combinations_shapes, 1)
        else:
            solution_count = self.sum_combination_products(solution_count,
                                                           first_n,
                                                           components_shape.combinations_shapes,
                                                           self._m_or_counters,
                                                           pmemo)

        ##############################################################
        # Uncrossed Independent Factors
        u_b_i = self._partitions.get_uncrossed_basic_independent_factors()
        for f in u_b_i:
            levels = list(filter(lambda l: not self._block.is_excluded_combination({f: l}), f.levels));
            if ind_factor_levels is not None:
                ind_factor_levels.append((f, levels))
            possibilities = pow(len(levels), first_n)
            solution_count *= possibilities
            components_shape.independent_shapes.append(possibilities)

        return solution_count

    def __count_preamble_solutions(self):
        if self._preamble_size == 0:
            return 1
        combos = 1
        for f in self._partitions.get_basic_factors():
            levels = list(filter(lambda l: not self._block.is_excluded_combination({f: l}), f.levels));
            self._basic_factor_levels.append((f, levels))
            combos *= len(levels)
        return pow(combos, self._preamble_size)

    def sum_combination_products(self, solution_count: int, first_n: int, shapes: List[int],
                                 m_or_counters: Union[int, List[int]],
                                 pmemo: PermutationMemo):
        if isinstance(m_or_counters, int):
            uniform_m = True
        elif all([m == m_or_counters[0] for m in m_or_counters]):
            uniform_m = True
        else:
            uniform_m = False
        if all([s == shapes[0] for s in shapes]) and uniform_m:
            # Every choice of `first_n` combinations produces the same number of possibilities
            # for source combinations in each trial
            return solution_count * pow(shapes[0], first_n)
        else:
            # Need to sum over the `solution_count` possible ways of picking `first_n` combinations,
            # getting the product of options across trials for each combination. Iterating over all
            # permutations is the slow way around, since the order within the permutation doesn't
            # matter (and we end up ignoring `_m_or_counters`), so there's room for improvement here:
            s = 0
            crossing_size = len(self._crossing_instances)
            for i in range(0, solution_count):
                permutation = compute_jth_prefix_of_permutations_with_copies(crossing_size, m_or_counters, first_n, i,
                                                                             pmemo)
                prod = 1
                for p in permutation:
                    prod *= shapes[p]
                s += prod
            return s




from typing import List, cast

from sweetpea._internal.sampling_strategy.base import Gen, SamplingResult
from sweetpea._internal.sampling_strategy.unigen import UniGen
from sweetpea._internal.block import Block


"""
This strategy relies CMSGen to sample from possible solutions in a way that
might be uniform, with without a firm guarantee of uniformity, so that the
lack of correlation would need to be checked independently.
"""
class CMSGen(Gen):
    # The CMSGen API is similar to Unigen, so we piggy-back on that implementation.

    @staticmethod
    def class_name():
        return 'CMSGen'

    @staticmethod
    def sample(block: Block, sample_count: int, min_search: bool=False) -> SamplingResult:
        return UniGen.sample(block, sample_count, min_search, use_cmsgen=True)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sweetpea",
    version="0.2.5",
    author="Annie Cherkaev, Ben Draut, Ahsan Sajjad Butt, Pierce Darragh, Matthew Flatt",
    author_email="annie.cherk@gmail.com, drautb@cs.utah.edu, ahsansbutt@hotmail.com, pierce.darragh@gmail.com, mflatt@cs.utah.edu",
    description="A language for synthesizing randomized experimental designs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sweetpea-org/sweetpea-py",
    packages=setuptools.find_packages(),
    install_requires=[
        'ascii-graph',
        'appdirs',
        'mypy',
        'networkx',
        'numpy',
        'pytest',
        'requests',
        'tqdm',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    test_suite='nose.collector',
    tests_require=['nose']
)



import pytest

from sweetpea import (
    Factor, Level, DerivedLevel, ElseLevel,
    MinimumTrials, WithinTrial, Transition,
    CrossBlock, Repeat,
    synthesize_trials, print_experiments,
    CMSGen, IterateSATGen, RandomGen
)

def check_consistent_solutions(block, expect):
    r_experiments  = synthesize_trials(block, 1000, RandomGen)
    i_experiments  = synthesize_trials(block, 1000, IterateSATGen)

    assert len(r_experiments) == expect
    assert len(i_experiments) == expect

    for e in r_experiments:
        assert e in i_experiments
    for e in i_experiments:
        assert e in r_experiments

def test_correct_solutions_with_plain_level_weight():
    color = Factor("color",  [Level("red", 1), "blue"])
    size = Factor("size",  ["big", Level("small", 2)])    

    constraints = []

    design       = [color, size]
    crossing     = [color, size]
    block        = CrossBlock(design, crossing, constraints)

    check_consistent_solutions(block, 180)

def test_correct_solutions_with_plain_level_weight_and_min_trials():
    color = Factor("color",  [Level("red", 1), "blue"])
    size = Factor("size",  ["big", Level("small", 2)])    

    constraints = [MinimumTrials(7)]

    design       = [color, size]
    crossing     = [color, size]
    block        = Repeat(CrossBlock(design, crossing, []),
                          constraints)

    check_consistent_solutions(block, 720)

@pytest.mark.slow
def test_correct_solutions_with_derived_level_weight():
    color = Factor("color",  ["red", "blue"])
    size = Factor("size",  ["big", "small"])

    look = Factor("look",
                  [DerivedLevel("bold",
                                WithinTrial(lambda c, s: c == "red" and s == "big",
                                            [color, size]),
                                2),
                   ElseLevel("plain")])

    constraints = [MinimumTrials(4)]
    
    design       = [color, size, look]
    crossing     = [look]

    check_consistent_solutions(CrossBlock(design, crossing, []), 9)
    check_consistent_solutions(Repeat(CrossBlock(design, crossing, [], constraints)), 36)

def test_correct_solutions_with_transition_level_weight():
    color = Factor("color",  ["red", "blue"])
    size = Factor("size",  ["big", "small"])

    look = Factor("look",
                  [DerivedLevel("bolden",
                                Transition(lambda c, s: c[0] == "red" and s[0] == "big",
                                           [color, size]),
                                2),
                   ElseLevel("weaken", 1)])

    constraints = [MinimumTrials(7)]
    
    design       = [color, size, look]
    crossing     = [look]
    
    check_consistent_solutions(CrossBlock(design, crossing, []), 36)
    check_consistent_solutions(Repeat(CrossBlock(design, crossing, []), constraints), 324)

def test_uncrossed_level_weight():
    color = Factor("color",  ["red", "blue"])
    size = Factor("size",  ["big", Level("small", 2)])

    design       = [color, size]
    crossing     = [color]
    block        = CrossBlock(design, crossing, [])

    check_consistent_solutions(CrossBlock(design, crossing, []), 18)
    check_consistent_solutions(Repeat(CrossBlock(design, crossing, []), [MinimumTrials(3)]), 108)
    
    experiments  = synthesize_trials(CrossBlock(design, crossing, []), 1000, RandomGen)
    totals = {}
    totals['big'] = 0
    totals['small'] = 0
    for e in experiments:
        for l in e['size']:
            totals[l] += 1

    assert totals['small'] == totals['big'] * 2
    assert set(experiments[0].keys()) == set(['color', 'size'])

def test_two_uncrossed_level_weight():
    same = Factor("same", "one")
    color = Factor("color",  ["red", Level("blue", 4)])
    size = Factor("size",  ["big", Level("medium", 2), Level("small", 3)])

    design       = [same, color, size]
    crossing     = [same]
    block        = CrossBlock(design, crossing, [])

    experiments  = synthesize_trials(CrossBlock(design, crossing, []), 1000, RandomGen)
    assert len(experiments) == 1000

    assert set(experiments[0].keys()) == set(['color', 'size', 'same'])

    totals = {}
    totals['red'] = 0
    totals['blue'] = 0
    totals['big'] = 0
    totals['medium'] = 0
    totals['small'] = 0
    for e in experiments:
        for l in e['size']:
            totals[l] += 1
        for l in e['color']:
            totals[l] += 1

    # These counts can fail, but with very low probability

    assert totals['small'] < totals['big'] * 4
    assert totals['small'] > totals['big'] * 2
    
    assert totals['medium'] < totals['big'] * 3
    assert totals['medium'] > totals['big'] * 1
    
    assert totals['blue'] < totals['red'] * 5
    assert totals['blue'] > totals['red'] * 3



"""This module provides functionality to test whether a CNF formula is
satisfiable.
"""


from ..cnf import CNF
from .tools.cryptominisat import cryptominisat_is_satisfiable
from .utility import save_cnf, temporary_cnf_file


__all__ = ['cnf_is_satisfiable']


def cnf_is_satisfiable(cnf: CNF) -> bool:
    """Determines whether the given CNF formula is satisfiable."""
    with temporary_cnf_file() as cnf_file:
        save_cnf(cnf_file, cnf)
        result = cryptominisat_is_satisfiable(cnf_file)
    if result:
        return True
    else:
        return False



from math import ceil, log
from tqdm import tqdm
import sys

from typing import List, cast

from sweetpea._internal.sampling_strategy.base import Gen, SamplingResult
from sweetpea._internal.block import Block
from sweetpea._internal.core import sample_uniform, CNF

"""
This strategy relies UniGen to sample uniformly from possible solutions.
"""
class UniGen(Gen):

    @staticmethod
    def class_name():
        return 'UniGen'

    @staticmethod
    def sample(block: Block, sample_count: int, min_search: bool=False, use_cmsgen=False) -> SamplingResult:

        backend_request = block.build_backend_request()
        if block.show_errors():
            return SamplingResult([], {})

        solutions = sample_uniform(
            sample_count,
            CNF(backend_request.get_cnfs_as_json()),
            backend_request.fresh - 1,
            block.variables_per_sample(),
            backend_request.get_requests_as_generation_requests(),
            use_docker=False,
            use_cmsgen=use_cmsgen)

        result = list(map(lambda s: Gen.decode(block, s.assignment), solutions))
        return SamplingResult(result, {})



"""This module provides combinatoric functionality."""


from typing import List, Tuple, Dict, Any, Union, Optional, cast
from math import factorial
from functools import reduce


def extract_components(sizes: List[int], n: int) -> List[int]:
    """Given a list of dimension sizes, and an integer less than the product of
    the sizes, this function will extract the component value for each
    dimension from the total.

    For example, if ``sizes`` is ``[3, 4, 2]``, there are 24 possible
    combinations of the three. If ``0 <= n < 24``, this function will return
    the selections for each dimension for that ``n``.
    """
    components = []
    for s in sizes:
        components.append(n % s)
        n //= s

    return components

##############################################################
# Finding combinations by index
#
# This is the simplest case: you want the `j`th unique combination of
# choices for `l` items, where each item has `n` possibilities. There
# are `pow(l, n)` such combinations, and we can treat the index as
# as base-n number.

def compute_jth_combination(l, n, j):
    """In a sequence of ``l`` items, where there are ``n`` choices for each
    item, this will compute the ``jth`` combination, out of all :math:`n^l`
    possibilities.
    """
    combination = [None] * l
    for k in range(l - 1, -1, -1):
        combination[k] = j % n
        j //= n

    return combination

##############################################################
# Finding combinations without replacement by index
#
# Treats the index as a number in the "combinatorial number system".

def compute_jth_combination_without_replacement(n, m, j) -> List[int]:
    """In a set of ``m`` items, where there are ``n`` choices for each
    item, this will compute the ``jth`` combination, out of all n-choose-m`
    possibilities.
    """
    combination = []
    while m > 0:
        c = m-1
        f_m = factorial(m)
        while c+1 < n and n_choose_m_given_m_factorial(c+1, m, f_m) <= j:
            c += 1
        j -= n_choose_m_given_m_factorial(c, m, f_m)
        combination.append(c)
        m -= 1

    return combination

def n_choose_m_given_m_factorial(n: int, m: int, f_m: int):
    if n < m:
        return 0
    if n == m:
        return 1
    p = 1
    h = n - m
    while n > h:
        p *= n
        n -= 1
    return p // f_m

def n_choose_m(n: int, m: int):
    return n_choose_m_given_m_factorial(n, m, factorial(m))

##############################################################
# Finding unique prefixes of permutations
#
# For permutations, the simple case is that you want the `j`th
# permutation of `n` elements, of which there are `factorial(n)`
# permutations.
#
# A case that's only slightly more complex: you want the `j`th
# distinct prefix of `m` elements in a `n`-element permutation. That's
# handled here, and it works due to the non-standard way we use the
# "inversion sequence" (so, it's not really an inversion sequence
# anymore, but it's closely related).

def compute_jth_permutation_prefix(n, m, j):
    """Compute the ``j``th prefix of ``m`` elements of a permutation of ``n`` elements.
    The ``j`` index is in terms of possible prefixes, not possible permutations."""
    inversion_sequence = compute_jth_inversion_sequence(n, m, j)
    return construct_permutation(inversion_sequence, n)

def compute_jth_inversion_sequence(n, m, j):
    """The are ``n!`` permutations of ``n`` elements. Each permutation can be
    uniquely identified by and constructed from its "inversion sequence". This
    function will compute the ``j``th inversion sequence for a set of ``n``
    elements.

    To deal with permutation prefixes, our ``j`` selects only the
    first m of n possible values. Also, we're going to change the
    usual meaning of an inversion sequence, using the values that we
    peel off (to construct a permutation) to mean how many unused
    numbers we skip over, because we want a prefix of the sequence to
    be able to represent any permutation prefix.

    Information on inversion sequences can be found:

    - `MathOnline <http://mathonline.wikidot.com/the-inversion-sequence-of-a-permutation>`__
    - "Introductory Combinatorics" by Richard A. Brualdi, ISBN 978-0136020400.

    """
    inversion = []
    for k in range(n, n-m, -1):
        result = j % k
        inversion.append(result)
        j //= k

    return inversion

def construct_permutation(inversion_sequence: List[int], orig_n: int) -> List[int]:
    """Given an inversion sequence, construct the permutation."""
    used = [False for i in range(orig_n)]
    permutation = [-1 for i in inversion_sequence]
    for n, skip in enumerate(inversion_sequence):
        idx = 0
        while used[idx]:
            idx += 1
        while skip > 0:
            if not used[idx]:
                skip -= 1
            idx += 1
        while used[idx]:
            idx += 1
        permutation[n] = idx
        used[idx] = True

    return permutation

##############################################################
# Finding permutations with repetitions
#
# How many permutations of `n` elements are there when each of the `n`
# elements is repeated exactly `m` times? It's straightforward to
# count those when we want the whole `n*m` sequence (see
# `count_permutation_with_copies`). Enumerating them is somewhat more
# challenging, but it's bascially a search built on counting.
# We build on this furrher below to find prefixes of such sequencs.
#
# The implementation below is based on
#   https://math.stackexchange.com/questions/3802978/permutation-with-repetition-index-conversion
#   https://stackoverflow.com/questions/24506460/algorithm-for-finding-multiset-permutation-given-lexicographic-index
#
# Initially, we're given:
#
#   q = noncomplex factor crossing
#
#   m = complex factor crossing size
#       [generalized to counter buckets when noncomplex factors are weighted]
#
# We have m copies of q combinations, where each copy is
# indistinguishable. The number of distinct permutations is not
# (q*m)!, because that would treat the copies as distinguishable. That
# is, for each of q combinations, (q*m)! includes m! cases that should
# be the same per combination, so overall it's K = (q*m)!/(m! ^ q)
# combinations.
#
# More generally, as implemented here, we can assign each of the q
# elements a number of copies c[i] (for i in 0 to q-1), and then there
# are
#
#  (c[0] + ... c[q-1])! / c[0]! * ... c[q-1]!
#
# permutations. To generate a permutation given an index from 0 to
# K-1, start with a counter array c initialized to m (or, more generally,
# a counter array is given):
#
#  - Pick the first non-empty counter and decrement it. Count
#    how many solutions would continue:
#
#    * If the index is less then that, you've found the first element
#      of the permutation; recur for the rest.
#
#   * If the index is more than that, reincrement the counter,
#     decrement the index by the number of solutions being skipped,
#     and try the next non-empty counter.
#
def _construct_permutation_with_copies(idx: int, q: int, fill_n: int, counters: List[int]) -> List[int]:
    sequence = cast(List[int], [])

    # assert sum(counters) == fill_n

    while len(sequence) < fill_n:
        i = 0
        while i < q:
            if counters[i] > 0:
                counters[i] -= 1
                n = count_remaining_permutations(counters)
                if idx >= n:
                    idx -= n
                    counters[i] += 1
                else:
                    sequence.append(i)
                    break
                i += 1
            else:
                i += 1
        assert i < q

    return sequence

def construct_permutation_with_copies(idx: int, q: int, m: int) -> List[int]:
    counters = [m for i in range(q)]
    return _construct_permutation_with_copies(idx, q, q*m, counters)

def construct_permutation_with_varying_copies(idx: int, q: int, counters: List[int]) -> List[int]:
    counters = counters[:]
    return _construct_permutation_with_copies(idx, q, sum(counters), counters)

def count_remaining_permutations(counters: List[int]) -> int:
    d = 1
    for c in counters:
        if c > 1:
            d = d * factorial(c)
    return factorial(sum(counters)) // d

def count_permutations_with_copies(q: int, m: int, first_n: int) -> int:
    n = q * m;
    if n == first_n:
        return factorial(n) // pow(factorial(m), q)
    else:
        # We want only the first_n of the trials, so count permutations
        # that only use the different ways of having first_n distributed
        # among the q possibilities.
        return cast(int, count_prefixes_of_permutations_with_copies(q, m, first_n, PermutationMemo()))

def count_permutations_with_varying_copies(q: int, counters: List[int], first_n: int) -> int:
    return cast(int, count_prefixes_of_permutations_with_copies(q, counters, first_n, PermutationMemo()))

##############################################################
# Finding unique prefixes of permutations with repetitions
#
# This is the general case. For a prefix of size `first_n` out of `q`
# elements repeated `m` times, just counting can be complex. One
# simple case is when `first_n` is no more than `m`; in that case,
# it's just a base-`q` number of `first_n` digits: `pow(q, first_n)`.
# Otherwise, the number of permutations depends on how you distribute
# `first_n` slots among the `q` choices, since using slots of one
# choice will reduce the number of distinct permutations. We use a
# dynamic-programming algorithm to count. The overall strategy is to
# try every allocation of `first_n` items to the `q` choices, and
# count the number of permutations of those `first_n` items (using the
# counter above, since we've reduced away the problem of prefixes).
#
# We start with a simple recursive algorithm, but that can easily blow
# the stack (thanks, Python!), so we also have one where we manage the
# continuation with our own stack. It's not much slower, but if the
# recursion wil be limited, we use the direct one, and otherwise we
# use the continuation one.
#
# We use the same algorithm to find the permutation at a given index,
# but always using the continuation one, so that we don't have to use
# a non-local escape in Python (although I guess an exception would
# work), but also so we don't have to implement that twice. If the
# memo table would cause us to skip over the item we want in the
# search, then we don't use it, which means that we navigate to the
# relevant choice of `first_n` items. We have to pass a multiplier
# down to that leaf, so it knows how many permutations total this leaf
# would account for; we keep that multiplier out of the memo table,
# because it would reduce sharing fatally.

class PermutationMemo():
    def __init__(self):
        self.memo = {}

def count_prefixes_of_permutations_with_copies(q: int, m_or_counters: Union[int, List[int]], first_n: int,
                                               pmemo: PermutationMemo) -> int:
    if isinstance(m_or_counters, list):
        return cast(int, k_prefixes_of_permutations_with_copies(q, m_or_counters, first_n, -1, pmemo))
    m = m_or_counters
    if first_n <= m:
        # Other techniques should produce the same result
        return pow(q, first_n)
    elif first_n < 100 and q < 100:
        return recur_count_prefixes_of_permutations_with_copies(q, m, first_n, pmemo)
    else:
        return cast(int, k_prefixes_of_permutations_with_copies(q, m, first_n, -1, pmemo))

def compute_jth_prefix_of_permutations_with_copies(q: int, m_or_counters: Union[int, List[int]], first_n: int, j: int,
                                                   pmemo: PermutationMemo) -> List[int]:
    if isinstance(m_or_counters, list):
        return cast(List[int], k_prefixes_of_permutations_with_copies(q, m_or_counters, first_n, j, pmemo))
    m = m_or_counters
    if first_n <= m:
        return compute_jth_combination(first_n, q, j)
    else:
        return cast(List[int], k_prefixes_of_permutations_with_copies(q, m, first_n, j, pmemo))

def recur_count_prefixes_of_permutations_with_copies(q: int, m: int, first_n: int,
                                                     pmemo: PermutationMemo) -> int:
    memo = pmemo.memo
    def recur(start_i: int, need_n: int, q: int, m: int) -> int:
        if need_n == 0:
            return 1
        elif start_i < q:
            if ((q - start_i) * m) >= need_n:
                combos = memo.get((start_i, need_n), None)
                if not combos:
                    combos = 0
                    for v in range(0, min(m, need_n)+1):
                        subs = recur(start_i+1, need_n - v, q, m)
                        combos += subs * count_interleavings(v, need_n)
                    memo[(start_i, need_n)] = combos
                return combos
            else:
                return 0
        else:
            return 0

    return recur(0, first_n, q, m)

def k_prefixes_of_permutations_with_copies(q: int, m_or_counters: Union[int, List[int]], first_n: int, find: int,
                                           pmemo: PermutationMemo) -> Union[int, List[int]]:
    # Dispatch on `m_or_counters` as a number (representing a uniform array) or an array:
    def available_after(start_i: int) -> int:
        nonlocal m_or_counters, q
        if isinstance(m_or_counters, list):
            return sum(m_or_counters[start_i:])
        else:
            return ((q - start_i) * m_or_counters)
    def available_at(start_i: int) -> int:
        nonlocal m_or_counters
        if isinstance(m_or_counters, list):
            return m_or_counters[start_i]
        else:
            return m_or_counters
    # Implements a find operation if `find` is > -1, otherwise just counts.
    # Based on `recur_count_prefixes_of_permutations_with_copies`, but explicitly
    # managing the continuation (so it can grows as deep as needed), and with two
    # extra arguments to enable generating the permutation instead of just counting.
    # Continuation frame ids:
    DoCount = 0
    DoNext = 1
    DoRecord = 2
    # Memo table for dynamic programming:
    memo = pmemo.memo
    # Value delivered to a continuation:
    value = 0
    # Our continuation, initialized with a call to count
    ks = cast(list, [(DoCount, (0,        # start_i
                                first_n,  # need_n
                                (),       # buckets: a cons-style list in reverse order, number of items allocated to a choice in `q`
                                1))])     # multiplier: so we can find a specific permutation
    while ks:
        k, state = ks.pop()
        if k == DoCount:
            start_i, need_n, buckets, multiplier = state
            if need_n == 0:
                # There's only one 0-length permutation
                value = 1
                if find > -1:
                    if find < multiplier:
                        # The permutation we want is within this bucket configuration
                        return _construct_permutation_with_copies(find, q, first_n, buckets_to_counters(buckets, q))
                    else:
                        find -= multiplier
            elif start_i >= q:
                value = 0
            elif available_after(start_i) < need_n:
                value = 0
            else:
                m_value = memo.get((start_i, need_n), None)
                if m_value and find > -1:
                    total = m_value * multiplier
                    if find < total:
                        m_value = None # explore this subtree, so we find the permutation
                    else:
                        find -= total
                if not m_value:
                    ks.append((DoNext, (0, start_i, need_n, 0, buckets, multiplier, 0)))
                    value = 0
                else:
                    value = m_value
        elif k == DoNext:
            v, start_i, need_n, count, buckets, multiplier, this_mult = state
            next_mult = count_interleavings(v, need_n)
            count += value * this_mult
            if v < min(available_at(start_i), need_n):
                ks.append((DoNext, (v+1, start_i, need_n, count, buckets, multiplier, next_mult)))
            else:
                ks.append((DoRecord, (v, start_i, need_n, count, next_mult)))
            ks.append((DoCount, (start_i+1, need_n - v, (v, buckets), next_mult * multiplier)))
        elif k == DoRecord:
            v, start_i, need_n, count, this_mult = state
            value = count + (value * this_mult)
            memo[(start_i, need_n)] = value
    return value

def count_interleavings(v: int, need_n: int) -> int:
    v_n = need_n - v
    # combinations with `v_n` elements, interleaved with `v` elements:
    return count_remaining_permutations([v_n, v])

# Converts the cons-style bucket representation of allocations
# to the `q` choices into a mutable-array counter representation
def buckets_to_counters(buckets: Any, q: int) -> List[int]:
    counters = [0 for i in range(q)]
    rev_buckets = cast(Any, ())
    while buckets != ():
        rev_buckets = (cast(Tuple[int, Any], buckets)[0], rev_buckets)
        buckets = cast(Tuple[int, Any], buckets)[1]
    i = 0;
    while rev_buckets != ():
        counters[i] = rev_buckets[0]
        rev_buckets = rev_buckets[1]
        i += 1
    return counters



import operator as op
import pytest

from itertools import permutations

from sweetpea import *
from sweetpea._internal.constraint import Reify
from sweetpea._internal.server import build_cnf
from acceptance import path_to_cnf_files, reset_expected_solutions


congruency            = Factor("congruency", ["congruent", "incongruent", "neutral"])
congruency_transition = Factor("congruency_transition", [
    DerivedLevel("con-con", Transition(lambda c: c[-1] == "congruent"   and c[0] == "congruent",   [congruency])),
    DerivedLevel("con-inc", Transition(lambda c: c[-1] == "congruent"   and c[0] == "incongruent", [congruency])),
    DerivedLevel("con-ntr", Transition(lambda c: c[-1] == "congruent"   and c[0] == "neutral",     [congruency])),
    DerivedLevel("inc-con", Transition(lambda c: c[-1] == "incongruent" and c[0] == "congruent",   [congruency])),
    DerivedLevel("inc-inc", Transition(lambda c: c[-1] == "incongruent" and c[0] == "incongruent", [congruency])),
    DerivedLevel("inc-ntr", Transition(lambda c: c[-1] == "incongruent" and c[0] == "neutral",     [congruency])),
    DerivedLevel("ntr-con", Transition(lambda c: c[-1] == "neutral"     and c[0] == "congruent",   [congruency])),
    DerivedLevel("ntr-inc", Transition(lambda c: c[-1] == "neutral"     and c[0] == "incongruent", [congruency])),
    ElseLevel("ntr-ntr")
])

@pytest.mark.parametrize('design', permutations([congruency, congruency_transition]))
def test_correct_solution_count_with_congruence_factor_but_unconstrained(design):
    crossing = [congruency]
    constraints = []

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100)

    assert len(experiments) == 6


def test_correct_solution_count_with_congruence_factor_but_unconstrained_cnf(design=[congruency, congruency_transition]):
    crossing = [congruency]
    constraints = list(map(Reify, design))

    block  = CrossBlock(design, crossing, constraints)
    cnf = build_cnf(block)

    if reset_expected_solutions:
        with open(path_to_cnf_files+'/test_correct_solution_count_with_congruence_factor_but_unconstrained.cnf', 'w') as f:
            f.write(cnf.as_unigen_string())
    with open(path_to_cnf_files+'/test_correct_solution_count_with_congruence_factor_but_unconstrained.cnf', 'r') as f:
        old_cnf = f.read()

    assert old_cnf == cnf.as_unigen_string()



"""This module provides non-uniform CNF sampling functionality through the
:func:`sample_non_uniform` function.

.. note::

    For compatibility with older testing mechanisms, there is also the
    :func:`sample_non_uniform_from_specification` function, which takes a
    :class:`.ProblemSpecification` as input. The :class:`.ProblemSpecification`
    can be generated with an appropriately formatted JSON file, which is how
    inputs were given in the original Haskell version of SweetPea Core. This
    function (and the associated classes) will likely be removed from SweetPea
    at some point.
"""


from pathlib import Path
from typing import List, Optional

from ..cnf import CNF
from .tools.cryptominisat import DEFAULT_DOCKER_MODE_ON, cryptominisat_solve
from .utility import GenerationRequest, ProblemSpecification, Solution, combine_and_save_cnf, temporary_cnf_file


__all__ = ['sample_non_uniform', 'sample_non_uniform_from_specification']


def sample_non_uniform(count: int,
                       initial_cnf: CNF,
                       fresh: int,
                       support: int,
                       generation_requests: List[GenerationRequest]
                       ) -> List[Solution]:
    """Samples solutions to a CNF problem non-uniformly. Produces ``count``
    solutions, each with a support set of length ``support``.
    """
    with temporary_cnf_file() as cnf_file:
        combine_and_save_cnf(cnf_file, initial_cnf, fresh, support, generation_requests)
        print("Running CryptoMiniSat...")
        solutions = compute_solutions(cnf_file, support, count)
        return [Solution(solution, 1) for solution in solutions]


def sample_non_uniform_from_specification(spec: ProblemSpecification) -> List[Solution]:
    """Samples solutions to a CNF problem non-uniformly, using a
    :class:`.ProblemSpecification`.

    .. note::

        This function exists for easier legacy compatibility from when
        SweetPea's input was given as JSON files. This should no longer be
        necessary.
    """
    return sample_non_uniform(spec.sample_count, spec.cnf, spec.fresh, spec.support, spec.requests)


def compute_solutions(filename: Path,
                      support: int,
                      count: int,
                      solutions: Optional[List[List[int]]] = None,
                      use_docker: bool = DEFAULT_DOCKER_MODE_ON
                      ) -> List[List[int]]:
    """Attempts to solve a CNF problem ``count`` times with CryptoMiniSAT. Each
    time a solution is generated, it is added to the problem file's header so
    new solutions may be generated. If at any point CryptoMiniSAT fails to
    generate a solution, execution terminates and the existing list of
    solutions will be returned.
    """
    # TODO: Implement iteratively instead of recursively.
    while True:
        if solutions is None:
            solutions = []
        if count == 0:
            return solutions
        solution = cryptominisat_solve(filename, use_docker)
        if not solution:
            return solutions
        solution = solution[:support]
        update_file(filename, solution)
        count -= 1
        solutions += [solution]


def update_file(filename: Path, solution: List[int]):
    """Updates a CNF file by adding a solution to the enclosed problem to the
    header. This allows CryptoMiniSAT to find additional (distinct) solutions
    to the problem.
    """

    def update_header(additional_clause_count: int, header: str) -> str:
        segments = header.strip().split()
        new_clause_count = int(segments[3]) + additional_clause_count
        return ' '.join(segments[:3] + [str(new_clause_count)])

    def add_clause_to_header(clause: str) -> str:
        return update_header(1, clause)

    lines = filename.read_text().strip().splitlines()
    updated_header = add_clause_to_header(lines[0])
    negated_solution = [-1 * var for var in solution]
    negated_solution_str = ' '.join(str(var) for var in negated_solution + [0])
    updated_lines = [updated_header] + lines[1:] + [negated_solution_str]
    updated_contents = '\n'.join(updated_lines)
    filename.write_text(updated_contents)



"""This module provides the fundamental functionality needed for SweetPea to
actually *do* anything. Primarily, this involves handling data representation
and making calls to external utilities for solving logic problems via
SAT-solving.


Data Representation
===================

SweetPea works by representing constraints on experimental designs as
`propositional logic formulas
<https://en.wikipedia.org/wiki/Propositional_formula>`_. These formulas are
converted into `conjunctive normal form
<https://en.wikipedia.org/wiki/Conjunctive_normal_form>`_ and are then passed
to an external SAT solver to either be solved or sampled.

Internally to :mod:`.core`, these formulas are represented as :class:`.CNF`
instances. These are comprised of :class:`Clauses <.Clause>`, which are in turn
comprised of :class:`Vars <.Var>`. The :class:`.Var`, :class:`.Clause`, and
:class:`.CNF` classes are very expressive and can easily be used to manipulate
advanced logic problems.


External Utilities
==================

Once the data is in a compatible formulaic representation, it must be shipped
to an external utility to be solved or sampled from. SweetPea Core makes use of
the following utilities:

  * `CryptoMiniSAT <https://github.com/msoos/cryptominisat>`_, an advanced
    incremental SAT solver.
  * `Unigen <https://github.com/meelgroup/unigen>`_, a state-of-the-art,
    almost-uniform sampler that uses CryptoMiniSAT.


Using Core
==========

There are only a few functions exported from :mod:`.core`, as well as a small
number of classes to support using those functions.

Functions
---------

  * :func:`~sweetpea.core.generate.is_satisfiable.cnf_is_satisfiable`
  * :func:`~sweetpea.core.generate.sample_non_uniform.sample_non_uniform`
  * :func:`~sweetpea.core.generate.sample_non_uniform.sample_non_uniform_from_specification`
  * :func:`~sweetpea.core.generate.sample_uniform.sample_uniform`
  * :func:`~sweetpea.core.generate.utility.combine_cnf_with_requests`

Classes
-------

  * :class:`~sweetpea.core.cnf.Var`
  * :class:`~sweetpea.core.cnf.Clause`
  * :class:`~sweetpea.core.cnf.CNF`
  * :class:`~sweetpea.core.generate.utility.AssertionType`
  * :class:`~sweetpea.core.generate.utility.GenerationRequest`
  * :class:`~sweetpea.core.generate.utility.Solution`
"""

from .cnf import Clause, CNF, Var
from .generate import (
    AssertionType, GenerationRequest, Solution,
    cnf_is_satisfiable, sample_non_uniform, sample_non_uniform_from_specification, sample_uniform,
    combine_cnf_with_requests
)
